from .adapter import Adapter
from .decorators import *
from .defer import Defer
from .singleton import singleton
from .signals import SignalBlocker
from .timestamp import time_since, TimeStamp
from .utils import *





class SignalBlocker:
    """ A context manager that blocks the signals of the provided widget.
    
    The signals are unblocked at the end of the with block.
    """
    def __init__(self, widget):
        self.widget = widget

    def __enter__(self):
        self.widget.blockSignals(True)

    def __exit__(self, *_):
        self.widget.blockSignals(False)


from qtstrap import *
from qtstrap.extras.code_editor import CodeEditor


class CodeLine(CodeEditor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setLineWrapMode(QTextEdit.NoWrap)
        self.setFixedHeight(28)

    def keyPressEvent(self, event:QKeyEvent):
        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            event.accept()
            return
        
        super().keyPressEvent(event)


from qtstrap import QObject, SignalInstance


class Adapter(QObject):
    """ A signal adapter that helps create disposable connections between objects. 
    
    A signal-based interface can be defined using an Adapter.

    Passing an existing Adapter when creating a new Adapter will automatically link all of
    the existing adapter's signals to the same-named signals on the new Adapter.

    This will allow some other object to connect to these signals for whatever purpose, and
    then simply delete the new Adapter object when it now longer wants to recieve signals.

    Technically, Qt Signals already have a .disconnect() method, but I've never gotten it work
    reliably. Using an Adapter essentially gives you a nuclear .disconnect().
    """
    def __init__(self, other=None):
        super().__init__()
        self._other = other
        if other is None:
            self._original = True
            return
        self._original = False

        for name in self._get_signals(other):
            getattr(other, name).connect(getattr(self, name).emit)

    def _get_signals(self, obj):
        signals = []
        for name in dir(obj):
            if name not in dir(QObject):
                if isinstance(getattr(obj, name), SignalInstance):
                    signals.append(name)
        return signals

    def __str__(self):
        s = ''
        if not self._original:
            s += ('inherited ')
        s += f"{self.__class__.__name__}(Adapter): <{', '.join(self._get_signals(self))}>"
        return s

    def __call__(self):
        return self.__class__(self)

    def adapter(self):
        return self.__class__(self)

    def kill(self):
        if self._other:
            for name in self._get_signals(self._other):
                getattr(self._other, name).disconnect(getattr(self, name).emit)

if __name__ == '__main__':
    from qtstrap import Signal

    class Example(QObject):
        class SignalInterface(Adapter):
            signal1 = Signal()
            signal2 = Signal(str)

        def __init__(self):
            super().__init__()
            self.interface = self.SignalInterface()

    ex = Example()
    adapter = ex.interface.adapter()

    print(ex.interface)
    print(adapter)


from qtstrap import *


class CodeEditor(QTextEdit):
    def __init__(self, *args, changed=None, model=None, highlighter=None, **kwargs):
        super().__init__(*args, **kwargs)

        set_font_options(self, {
            'setFamily': 'Courier',
            'setStyleHint': QFont.Monospace,
            'setFixedPitch': True,
        })
        
        if qtpy.QT5:
            self.setTabStopWidth(QFontMetricsF(self.font()).width(' ') * 4)
        elif qtpy.QT6:
            self.setTabStopWidth(QFontMetricsF(self.font()).tightBoundingRect(' ').width() * 4)
        if highlighter:
            self.syntax = highlighter(self)
        
        if changed:
            self.textChanged.connect(changed)

        self.model = model

        self.completer = QCompleter(self.model, self)
        self.completer.setWidget(self)
        self.completer.popup().setMinimumWidth(150)
        self.completer.setCompletionMode(QCompleter.UnfilteredPopupCompletion)

        self.completer.activated.connect(self.insert_completion)

    def insert_completion(self, completion):
        tc = self.textCursor()
        extra = len(completion) - len(self.completer.completionPrefix())
        tc.movePosition(QTextCursor.Left)
        tc.movePosition(QTextCursor.EndOfWord)
        tc.insertText(completion[-extra:])
        self.setTextCursor(tc)

    def text_under_cursor(self):
        tc = self.textCursor()
        tc.select(QTextCursor.WordUnderCursor)
        return tc.selectedText()
        
    def keyPressEvent(self, event:QKeyEvent):
        force_popup = False
        if event.modifiers() == Qt.ControlModifier:
            if event.key() == Qt.Key_Space:
                force_popup = True

        keys = [Qt.Key_Enter, Qt.Key_Return, Qt.Key_Escape, Qt.Key_Tab, Qt.Key_Backtab]
        if self.completer.popup().isVisible():
            if event.key() in keys:
                event.ignore()
                return

        braces = {
            '"': '"',
            "'": "'",
            '{': '}',
            '(': ')',
            '<': '>',
            '[': ']',
            '|': '|',
            '`': '`',
        }

        cur = self.textCursor()
        if cur.hasSelection():
            if event.text() in braces:
                start = cur.selectionStart()
                end = cur.selectionEnd()
                cur.clearSelection()
                cur.setPosition(end)
                cur.insertText(braces[event.text()])
                cur.setPosition(start)
                cur.insertText(event.text())
                cur.setPosition(start + 1)
                cur.setPosition(end + 1, QTextCursor.KeepAnchor)
                self.setTextCursor(cur)
                return

        super().keyPressEvent(event)

        prefix = self.text_under_cursor()
        if event.key() == Qt.Key_Period:
            force_popup = True
        if prefix or force_popup:
            if self.model:
                self.model.set_prefix(prefix, self.textCursor())
            self.completer.setCompletionPrefix(prefix)
            index = self.completer.completionModel().index(0, 0)
            self.completer.popup().setCurrentIndex(index)

            cr = self.cursorRect()
            cr.setWidth(self.completer.popup().sizeHintForColumn(0) + self.completer.popup().verticalScrollBar().sizeHint().width())
            self.completer.complete(cr)
        else:
            self.completer.popup().hide()

    def text(self):
        return self.toPlainText()


class Defer:
    """ A context manager that emulates the defer keyword from other languages.

    The deferred thing can be any callable, and arbitrary args and kwargs will be preserved
    and passed to the thing during __exit__().
    """

    def __init__(self, thing, *args, **kwargs):
        self.thing = thing
        self.args = args
        self.kwargs = kwargs

    def __enter__(self): ...

    def __exit__(self, *_):
        self.thing(*self.args, **self.kwargs)


if __name__ == '__main__':
    print('first')
    with Defer(lambda: print('fourth')):
        with Defer(print, 'third'):
            print('second')



from time import time


def time_since(timestamp):
    return (time() - timestamp)


class TimeStamp:
    def __init__(self):
        self._time = time()

    def time_since(self, timestamp=None):
        if timestamp is None:
            return time() - self._time
            
        if type(timestamp) == TimeStamp():
            return time() - timestamp._time

        return time() - timestamp

    def update(self):
        self._time = time()

    def __repr__(self):
        return str(self._time)